@@ -1,17 +1,22 @@
+from operator import ge
 from koil.helpers import unkoil
 from koil.koil import Koil
 import inspect
 from typing import Callable, Type, TypeVar
 from koil.vars import current_loop
+import logging
 
 T = TypeVar("T")
 
 
+logger = logging.getLogger(__name__)
+
+
 def koilable(
     fieldname: str = "__koil",
     add_connectors: bool = False,
     koil_class: Type[Koil] = Koil,
-    **koilparams
+    **koilparams,
 ) -> Callable[[Type[T]], Type[T]]:
     """
     Decorator to make an async generator koilable.
@@ -30,10 +35,19 @@ def koilable(
             cls.__aexit__
         ), "__aexit__ must be a coroutine"
 
+        def ___get_koiled_loop(self):
+            return getattr(self, fieldname)
+
         def koiled_enter(self, *args, **kwargs):
-            if current_loop.get() is None:
-                if getattr(self, fieldname, None) is None:
-                    setattr(self, fieldname, koil_class(**koilparams))
+            potential_koiled_loop = current_loop.get()
+            if potential_koiled_loop is not None:
+                assert (
+                    getattr(self, fieldname, None) is None
+                ), f"You cannot enter a koil loop inside another koil loop. Do no set explicitly a Koil in {cls}. Found koiled loop {potential_koiled_loop}"
+            else:
+                setattr(
+                    self, fieldname, koil_class(creating_instance=self, **koilparams)
+                )
                 getattr(self, fieldname).__enter__()
 
             return unkoil(self.__aenter__, *args, **kwargs)
@@ -43,14 +57,10 @@ def koilable(
             koil = getattr(self, fieldname, None)
             if koil is not None:
                 koil.__exit__(None, None, None)
-            setattr(self, fieldname, None)
+                setattr(self, fieldname, None)
 
         def disconnect(self):
-            unkoil(self.__aexit__, None, None, None)
-            koil = getattr(self, fieldname, None)
-            if koil is not None:
-                koil.__exit__(None, None, None)
-            setattr(self, fieldname, None)
+            return self.__exit__(None, None, None)
 
         async def adisconnect(self):
             return await self.__aexit__(None, None, None)
@@ -65,6 +75,7 @@ def koilable(
             cls.aconnect = aconnect
             cls.disconnect = disconnect
             cls.connect = koiled_enter
+            cls.__get_koiled_loop = ___get_koiled_loop
 
         return cls
 
