@@ -3,20 +3,30 @@ from asyncio.log import logger
 import threading
 
 import janus
-from koil.errors import KoilError, KoilStopIteration, ThreadCancelledError
+from koil.errors import (
+    CancelledError,
+    KoilError,
+    KoilStopIteration,
+    ThreadCancelledError,
+)
+from koil.task import KoilFuture, KoilRunner
 from koil.utils import run_threaded_with_context
 from koil.vars import *
-from koil.task import KoilTask
 import time
 
 
-def create_task(coro, *args, **kwargs) -> KoilTask:
-    """Create a coil task from a coroutine"""
-    return coro(*args, **kwargs, as_task=True)
+def unkoil_gen(iterator, *args, **kwargs):
+    try:
+        loop0 = asyncio.events.get_running_loop()
+        # We are running inside a loop
+        # TODO: CHeck if iterator self has a loop attribute, if so, check if it is the same loop
+        return iterator(*args, **kwargs)
+    except RuntimeError:
+        pass
 
-
-def unkoil_gen(iterator, *args, timeout=None, **kwargs):
     loop = current_loop.get()
+    assert loop, "No koiled loop found"
+
     cancel_event = current_cancel_event.get() or threading.Event()
 
     if loop.is_closed():
@@ -30,12 +40,15 @@ def unkoil_gen(iterator, *args, timeout=None, **kwargs):
 
     ait = iterator(*args, **kwargs).__aiter__()
     res = [False, False]
-    next_args = ()
+    next_args = None
 
-    async def next_on_ait(next_args):
+    async def next_on_ait(inside_args):
         try:
             try:
-                obj = await ait.__anext__(*next_args if next_args else ())
+                if inside_args:
+                    obj = await ait.__anext__(*inside_args)
+                else:
+                    obj = await ait.__anext__()
                 return [False, obj]
             except StopAsyncIteration:
                 return [True, None]
@@ -43,9 +56,7 @@ def unkoil_gen(iterator, *args, timeout=None, **kwargs):
             return [False, e]
 
     while True:
-        res = run_threaded_with_context(
-            next_on_ait(next_args), loop, cancel_event=cancel_event
-        )
+        res = run_threaded_with_context(next_on_ait, loop, cancel_event, next_args)
         x, context = res.result()
         done, obj = x
         if done:
@@ -59,13 +70,12 @@ def unkoil_gen(iterator, *args, timeout=None, **kwargs):
         next_args = yield obj
 
 
-def unkoil(coro, *args, timeout=None, as_task=False, ensure_koiled=False, **kwargs):
+def unkoil(coro, *args, **kwargs):
     try:
         loop = asyncio.events.get_running_loop()
-        if loop == current_loop.get():
-            raise NotImplementedError(
-                "Calling sync() from within a running loop, you need to await the coroutine"
-            )
+        return coro(
+            *args, **kwargs
+        )  # We are running in an event loop so we can just return the coroutine
 
     except RuntimeError:
         pass
@@ -88,18 +98,11 @@ def unkoil(coro, *args, timeout=None, as_task=False, ensure_koiled=False, **kwar
                 newcontext = contextvars.copy_context()
                 return x, newcontext
 
-            if as_task:
-                taskclass = current_taskclass.get()
-                assert taskclass is not None, "No task class set"
-                return taskclass(
-                    coro, preset_args=args, preset_kwargs=kwargs, loop=loop
-                )
-
             co_future = asyncio.run_coroutine_threadsafe(passed_with_context(), loop)
             while not co_future.done():
                 time.sleep(0.01)
                 if cancel_event and cancel_event.is_set():
-                    raise Exception("Task was cancelled")
+                    raise ThreadCancelledError("Task was cancelled")
 
             x, newcontext = co_future.result()
 
@@ -111,22 +114,10 @@ def unkoil(coro, *args, timeout=None, as_task=False, ensure_koiled=False, **kwar
         except KeyboardInterrupt:
             print("Grace period triggered?")
             raise
-    else:
-        if ensure_koiled:
-            raise RuntimeError("No loop set and ensure_koiled was set to True")
 
-        if as_task:
-            raise RuntimeError(
-                """No loop is running. That means you cannot have this run as a task. Try providing a loop by entering a Koil() context.
-                """
-            )
-
-        logger.warn(
-            "You used unkoil without a governing Koil in the context, this is not recommended. We will now resort to run asyncio.run()"
-        )
-
-        future = coro(*args, **kwargs)
-        asyncio.run(future)
+    raise NotImplementedError(
+        f"You need to be in a Koil() context to use sync() {coro} {loop}"
+    )
 
 
 async def run_spawned(
@@ -149,7 +140,6 @@ async def run_spawned(
             assert loop0 is loop, "Loop is not the same"
         else:
             loop = loop0
-            current_taskclass.set(KoilTask)
     except RuntimeError:
         loop = current_loop.get()
 
@@ -159,7 +149,6 @@ async def run_spawned(
     def wrapper(sync_args, sync_kwargs, loop, cancel_event, context):
         if pass_loop:
             current_loop.set(loop)
-            current_taskclass.set(KoilTask)
 
         current_cancel_event.set(cancel_event)
 
@@ -215,7 +204,6 @@ async def iterate_spawned(
             assert loop0 is loop, "Loop is not the same"
         else:
             loop = loop0
-            current_taskclass.set(KoilTask)
     except RuntimeError:
         loop = current_loop.get()
 
@@ -237,7 +225,6 @@ async def iterate_spawned(
     ):
         if pass_loop:
             current_loop.set(loop)
-            current_taskclass.set(KoilTask)
 
         current_cancel_event.set(cancel_event)
 
@@ -302,3 +289,11 @@ async def iterate_spawned(
                 raise finish_condition
             except KoilStopIteration:
                 break
+
+
+def create_task(coro, *args, **kwargs) -> KoilFuture:
+    return KoilRunner(coro, preset_args=args, preset_kwargs=kwargs).run()
+
+
+def create_runner(coro, *args, **kwargs) -> KoilRunner:
+    return KoilRunner(coro, preset_args=args, preset_kwargs=kwargs)
